/*****************************************************************************
** FILE IDENTIFICATION
**
**   Name:	   projections.cpp         Projection data classes
**   Programmer:   Kevin Rosenberg
**   Date Started: Aug 84
**
**  This is part of the CTSim program
**  Copyright (C) 1983-2000 Kevin Rosenberg
**
**  $Id: projections.cpp,v 1.2 2000/06/19 17:58:20 kevin Exp $
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License (version 2) as
**  published by the Free Software Foundation.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

#include "ct.h"


/* NAME
 *    Projections		Constructor for projections matrix storage 
 *
 * SYNOPSIS
 *    proj = projections_create (filename, nView, nDet)
 *    Projections& proj		Allocated projections structure & matrix
 *    int nView			Number of rotated view
 *    int nDet			Number of detectors
 *
 */

Projections::Projections (const Scanner& scanner)
  : m_projData(0)
{
  initFromScanner (scanner);
}


Projections::Projections (const int nView, const int nDet)
  : m_projData(0)
{
  init (nView, nDet);
}

Projections::Projections (void)
  : m_projData(0)
{
  init (0, 0);
}

Projections::~Projections (void)
{
  deleteProjData();
}


void
Projections::init (const int nView, const int nDet)
{
  m_nView = nView;
  m_nDet = nDet;
  newProjData ();
}

void
Projections::initFromScanner (const Scanner& scanner)
{
  deleteProjData();
  init (scanner.nView(), scanner.nDet());

  m_phmLen = scanner.phmLen();
  m_rotInc = scanner.rotInc();
  m_detInc = scanner.detInc();
  m_rotStart = 0;
  m_detStart =  -scanner.radius() + (scanner.detInc() / 2);
  m_phmLen = scanner.phmLen();
}

void
Projections::setNView (int nView)  // used by MPI to reduce # of views
{
  deleteProjData();
  init (nView, m_nDet);
}

// NAME
// newProjData

void 
Projections::newProjData (void)
{
  if (m_projData)
    sys_error(ERR_WARNING, "m_projData != NULL [newProjData]");

  if (m_nView > 0 && m_nDet) {
    m_projData = new DetectorArray* [m_nView];

    for (int i = 0; i < m_nView; i++)
      m_projData[i] = new DetectorArray (m_nDet);
  }
}


/* NAME
 *   projections_free			Free memory allocated to projections
 *
 * SYNOPSIS
 *   projections_free(proj)
 *   Projections& proj				Projectionss to be deallocated
 */

void 
Projections::deleteProjData (void)
{
  if (m_projData != NULL) {
    for (int i = 0; i < m_nView; i++)
      delete m_projData[i];

    delete m_projData;
    m_projData = NULL;
  }
}


/* NAME
 *    projections_write_header         Write data header for projections file
 *
 */

bool 
Projections::headerWrite (void)
{
  kuint32 _hsize = m_headerSize;
  kuint32 _nView = m_nView;
  kuint32 _nDet = m_nDet;
  kuint32 _geom = m_geometry;
  kuint32 _remarksize = m_remark.length();
  kfloat64 _calcTime = m_calcTime;
  kfloat64 _rotStart = m_rotStart;
  kfloat64 _rotInc = m_rotInc;
  kfloat64 _detStart = m_detStart;
  kfloat64 _detInc = m_detInc;
  kfloat64 _phmLen = m_phmLen;
  
  if (m_fd < 0) {
    sys_error (ERR_SEVERE, "m_fd < 0 [projections_write_header]");
    return false;
  }
    
  if (lseek (m_fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
    sys_error (ERR_WARNING, "Error seeking to beginning of fiel [projections_write_header]");
    return false;
  }

  if (! write_nint32 (&_hsize, m_fd) ||
      ! write_nint32 (&_nView, m_fd) ||
      ! write_nint32 (&_nDet, m_fd) ||
      ! write_nint32 (&_geom, m_fd) ||
      ! write_nfloat64 (&_calcTime, m_fd) ||
      ! write_nfloat64 (&_rotStart, m_fd) ||
      ! write_nfloat64 (&_rotInc, m_fd) ||
      ! write_nfloat64 (&_detStart, m_fd) ||
      ! write_nfloat64 (&_detInc, m_fd) ||
      ! write_nfloat64 (&_phmLen, m_fd) ||
      ! write_nint32 (&_remarksize, m_fd) ||
      (::write (m_fd, m_remark.c_str(), _remarksize) != _remarksize)) {
    sys_error (ERR_SEVERE, "Error writing header information [projections_write_header] %ld", _remarksize);
    return false;
  }
  m_headerSize = _hsize = lseek(m_fd, (off_t) 0, SEEK_CUR);
  if ((lseek(m_fd, 0, SEEK_SET) != 0) || ! write_nint32 (&_hsize, m_fd)) {
    sys_error (ERR_SEVERE, "Error writing header information [projections_write_header]");
    return false;
  }
  
  return true;
}


/* NAME
 *    projections_read_header         Read data header for projections file
 *
 */

bool
Projections::headerRead (void)
{
  kuint32 _hsize;
  kuint32 _nView;
  kuint32 _nDet;
  kuint32 _geom;
  kuint32 _remarksize = 0;
  kfloat64 _calcTime;
  kfloat64 _rotStart;
  kfloat64 _rotInc;
  kfloat64 _detStart;
  kfloat64 _detInc;
  kfloat64 _phmLen;
  
  if (m_fd < 0) {
    sys_error (ERR_SEVERE, "m_fd < 0 [projections_read_header]");
    return false;
  }
    
  if (lseek (m_fd, (off_t) 0, SEEK_SET) != (off_t) 0) {
    sys_error (ERR_WARNING, "Error seeking to beginning of field [projections_read_header]");
    return false;
  }

  off_t testPos;
  if (! read_nint32 (&_hsize, m_fd)) {
    sys_error (ERR_SEVERE, "Error reading header size , _hsize=%d [projections_read_header]", _hsize);
    return false;
  }
  if ( ! read_nint32 (&_nView, m_fd) ||
      ! read_nint32 (&_nDet, m_fd) ||
      ! read_nint32 (&_geom, m_fd) ||
      ! read_nfloat64 (&_calcTime, m_fd) ||
      ! read_nfloat64 (&_rotStart, m_fd) ||
      ! read_nfloat64 (&_rotInc, m_fd) ||
      ! read_nfloat64 (&_detStart, m_fd) ||
      ! read_nfloat64 (&_detInc, m_fd) ||
      ! read_nfloat64 (&_phmLen, m_fd) ||
       ! read_nint32 (&_remarksize, m_fd))
      {
	  sys_error (ERR_SEVERE, "Error reading header information , _remarksize=%d [projections_read_header]", _remarksize);
	  return false;
      }

  char remarkStorage[_remarksize+1];
  if (::read (m_fd, &remarkStorage, _remarksize) != _remarksize) {
    sys_error (ERR_SEVERE, "Error reading remark, _remarksize = %d", _remarksize);
    return false;
  }
  remarkStorage[_remarksize] = 0;
  m_remark = remarkStorage;

  off_t _hsizeread = lseek(m_fd, 0, SEEK_CUR);
  if (_hsizeread != _hsize) {
      sys_error (ERR_WARNING, "File header size read %ld != file header size stored %ld [read_projections_header]\n_remarksize=%ld", (long int) _hsizeread, _hsize, _remarksize);
      return false;
  }
  
  m_headerSize = _hsize;
  m_nView = _nView;
  m_nDet = _nDet;
  m_geometry = _geom;
  m_calcTime = _calcTime;
  m_rotStart = _rotStart;
  m_rotInc = _rotInc;
  m_detStart = _detStart;
  m_detInc = _detInc;
  m_phmLen = _phmLen;

  return true;
}


bool
Projections::read (const char* filename)
{
  if ((m_fd = open(filename, OPEN_RDONLY | O_BINARY)) < 0) {
    sys_error (ERR_SEVERE, "Error opening projections file %s for input [open_projections]", filename);
    return false;
  }

  if (! headerRead()) {
    sys_error (ERR_SEVERE, "Error reading projections file %s [open_projections]", filename);
    return false;
  }

  deleteProjData ();
  newProjData ();

  for (int i = 0; i < m_nView; i++) {
    if (! detarrayRead (*m_projData[i], i))
      break;
  }

  close (m_fd);
  m_fd = -1;

  return true;
}


bool
Projections::write (const char* filename)
{
    if ((m_fd = open (filename, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY 
#ifndef _WIN32
			, S_IWRITE | S_IREAD | S_IRGRP | S_IROTH
#endif
			)) < 0) {
      sys_error (ERR_SEVERE, "Error opening file %s for output [projections_create]", filename);
      return false;
    }
    if (! headerWrite ()) {
      sys_error(ERR_SEVERE, "Error writing to file %s [create_projections]", filename);
      return false;
    }

    headerWrite ();

  if (m_projData != NULL) {
    for (int i = 0; i < m_nView; i++) {
      if (! detarrayWrite (*m_projData[i], i))
	break;
    }
  }

  close (m_fd);
  m_fd = -1;

  return true;
}

/* NAME
 *   detarrayRead		Read a Detector Array structure from the disk
 *
 * SYNOPSIS
 *   detarrayRead (proj, darray, view_num)
 *   DETARRAY *darray		Detector array storage location to be filled
 *   int      view_num		View number to read
 */

bool
Projections::detarrayRead (DetectorArray& darray, const int iview)
{
  const int detval_bytes = darray.nDet() * sizeof(kfloat32);
  const int detheader_bytes = sizeof(kfloat64) /* view_angle */ + sizeof(kint32) /* nDet */;
  const int view_bytes = detheader_bytes + detval_bytes;
  const off_t start_data = m_headerSize + (iview * view_bytes);
  DetectorValue* detval_ptr = darray.detValues();  
  kfloat64 view_angle;
  kuint32 nDet;

  if (m_fd < 0) { 
    sys_error (ERR_SEVERE, "Tried to read detector array on closed file [detarrayRead]"); 
    return false; 
  } 

  if (lseek(m_fd, start_data, SEEK_SET) != start_data) {
    sys_error (ERR_SEVERE, "Error seeking detectory array [detarrayRead]"); 
    return false; 
  }

  if (! read_nfloat64 (&view_angle, m_fd) ||
      ! read_nint32 (&nDet, m_fd)) {
      sys_error (ERR_SEVERE, "Error reading detector array [detarrayRead]");
      return false;
  }
  darray.setViewAngle (view_angle);
  //  darray.setNDet ( nDet);
  
  for (int i = 0; i < nDet; i++) {
      kfloat32 detval;
      if (! read_nfloat32 (&detval, m_fd)) {
	  sys_error (ERR_SEVERE, "Error reading detector array [detarrayRead]");
	  return false;
      }
      detval_ptr[i] = detval;
  }

  return true;
}


/* NAME
 *   detarrayWrite			Write detector array data to the disk
 *
 * SYNOPSIS
 *   detarrayWrite (darray, view_num)
 *   DETARRAY *darray			Detector array data to be written
 *   int      view_num			View number to write
 *
 * DESCRIPTION
 *       This routine writes the detarray data from the disk sequentially to
 *    the file that was opened with open_projections().  Data is written in
 *    binary format.
 */

bool
Projections::detarrayWrite (const DetectorArray& darray, const int iview)
{
  const int detval_bytes = darray.nDet() * sizeof(float);
  const int detheader_bytes = sizeof(kfloat64) /* view_angle */ + sizeof(kint32) /* nDet */;
  const int view_bytes = detheader_bytes + detval_bytes;
  const off_t start_data = m_headerSize + (iview * view_bytes);
  const DetectorValue* const detval_ptr = darray.detValues();  
  kfloat64 view_angle = darray.viewAngle();
  kuint32 nDet = darray.nDet();
  
  if (m_fd < 0) {
    sys_error (ERR_SEVERE, "Tried to write detector array with no file open [detarrayWrite]");
    return false;
  }

  if (lseek(m_fd, start_data, SEEK_SET) != start_data) {
      sys_error (ERR_SEVERE, "Error seeking detectory array [detarrayWrite]");
      return false;
  }

  if ( ! write_nfloat64 (&view_angle, m_fd)) {
      sys_error (ERR_SEVERE, "Error writing detector array [detarrayWrite]");
      return false;
  }
  if ( ! write_nint32 (&nDet, m_fd)) {
      sys_error (ERR_SEVERE, "Error writing detector array [detarrayWrite]");
      return false;
  }
  for (int i = 0; i < nDet; i++) {
      kfloat32 detval = detval_ptr[i];
      if ( ! write_nfloat32 (&detval, m_fd)) {
	  sys_error (ERR_SEVERE, "Error writing detector array [detarrayWrite]");
	  return false;
      }
  }

  return true;
}


/* NAME
 *   prt_projections			Print projections data
 *
 * SYNOPSIS
 *   prt_projections (proj)
 *   Projections& proj			Projection data to be printed
 */

void
Projections::printProjectionData (void)
{
  printf("Projections Print\n\n");
  printf("Description: %s\n", m_remark.c_str());
  printf("nView = %d  nDet = %d\n", m_nView, m_nDet);
  printf("rotStart = %8.4f   rotInc = %8.4f\n", m_rotStart, m_rotInc);
  printf("detStart = %8.4f   detInc = %8.4f\n", m_detStart, m_detInc);

  if (m_projData != NULL) {
      for (int ir = 0; ir < m_nView; ir++) {
	  DetectorValue* detval = m_projData[ir]->detValues();
	  for (int id = 0; id < m_projData[ir]->nDet(); id++)
	      printf("%8.4f  ", detval[id]);
	  printf("\n");
      }
  }
}

void 
Projections::printScanInfo (void) const
{
  printf ("Number of detectors: %d\n", m_nDet);
  printf ("    Number of views: %d\n", m_nView);
  printf ("             Remark: %s\n", m_remark.c_str());
  printf ("             phmLen: %f\n", m_phmLen);
  printf ("           detStart: %f\n", m_detStart);
  printf ("             detInc: %f\n", m_detInc);
  printf ("           rotStart: %f\n", m_rotStart);
  printf ("             rotInc: %f\n", m_rotInc);
}


